/*
 * Filename:    register.cpp
 *
 * This file handles installation/uninstallation startup and shutdown of the
 * nullspi It's done by DllRegister and UnRegister functions exported by
 * nullspi.dll, should use regdll32.exe to register it
 *
 * Author: xwh@cs.cornell.edu
 */
#define UNICODE

#include <string.h>
#include <tchar.h>
#include <malloc.h>
#include "nullspi.h"
#include <stdlib.h>

#define NULL_PROVIDER_NAME L"NULL_PROVIDER"

// globals

long             thread_count ;
HMODULE          g_hinst ;
CRITICAL_SECTION gInitCriticalSection;
WSPUPCALLTABLE   gupCallTable;
INT              gCatalog;

INT InstallMyProvider( PDWORD CatalogId)
{
	WSAPROTOCOL_INFOW  proto_info;
	int               install_result;
	int               install_error;
	WCHAR             DllPath	[_MAX_PATH];
	char              tempbuf[_MAX_PATH]; tempbuf[0]='\0';
	GUID nullProviderId = {0,0,0,{0,0,0,0,0,0,0,0}};

	// Create a PROTOCOL_INFO to install for our provider DLL.

	// this flag specifies the "service" provided
	// I will set it to be   XP1_MESSAGE_ORIENTED later
	proto_info.dwServiceFlags1 = 0;
	proto_info.dwServiceFlags2 = 0;
	proto_info.dwServiceFlags3 = 0;
	proto_info.dwServiceFlags4 = 0;

	proto_info.dwProviderFlags = PFL_HIDDEN;  // kind of important

	proto_info.ProviderId      = nullProviderId;
	proto_info.dwCatalogEntryId = 0;   // filled in by system
	proto_info.ProtocolChain.ChainLen = BASE_PROTOCOL; 
	proto_info.iVersion = 0;
	proto_info.iAddressFamily = AF_ATM; // important
	proto_info.iMaxSockAddr = 16;
	proto_info.iMinSockAddr = 16;
	proto_info.iSocketType = SOCK_DGRAM;  // important
	proto_info.iProtocol = IPPROTO_UDP;   // important  
	proto_info.iProtocolMaxOffset = 0;
	proto_info.iNetworkByteOrder = BIGENDIAN;
	proto_info.iSecurityScheme = SECURITY_PROTOCOL_NONE;
	proto_info.dwMessageSize = 64*1024;  // 64k
	proto_info.dwProviderReserved = 0xdeadbabe;


	wcscpy( proto_info.szProtocol, NULL_PROVIDER_NAME);

	GetModuleFileName( g_hinst, DllPath, _MAX_PATH);
	WideCharToMultiByte(CP_ACP, 0, DllPath, wcslen ( DllPath) +1 ,
			tempbuf, _MAX_PATH , NULL, NULL);

	install_result = WSCInstallProvider(
			&NullSPIGuid,

			DllPath,                      // lpszProviderDllPath
			& proto_info,                 // lpProtocolInfoList
			1,                            // dwNumberOfEntries
			& install_error);             // lpErrno

	// this entry is filled by the WSCInstallProvider()
	*CatalogId = proto_info.dwCatalogEntryId;
	return(install_result);
} 

HRESULT DllRegisterServer(void)
{
	// this is the code that is called by regsvr32.exe.
	// so the installation starts here

	WCHAR textBuf[_MAX_PATH];
	DWORD CatalogId;
	INT   RetValue = NOERROR;

	wsprintf(textBuf,L"Press OK to contiue installation");
	MessageBox(NULL, textBuf, L"NULL SPI Installation", MB_OK);
	RetValue = InstallMyProvider( & CatalogId);

	return  RetValue;
}

HRESULT DllUnregisterServer(void)
{
	// this is the code that is called by regsvr32.exe for
	// uninstallaion
	// so the uninstallation starts here
	//
	INT RetValue = NOERROR;
	INT error;

	RetValue = WSCDeinstallProvider(&NullSPIGuid, &error) ;
	return RetValue;
}

BOOL WINAPI DllMain(
		IN HINSTANCE hinstDll,
		IN DWORD fdwReason,
		LPVOID lpvReserved
									 )
{

	switch (fdwReason) {

		case DLL_PROCESS_ATTACH:
			// DLL is attaching to the address space of the current process.
			InitializeCriticalSection(&gInitCriticalSection);
			g_hinst = hinstDll;
			thread_count = 0;

			break;
		case DLL_THREAD_ATTACH:
			// A new thread is being created in the current process.
			InterlockedIncrement ( & thread_count);

			break;
		case DLL_THREAD_DETACH:
			// A thread is exiting cleanly.
			InterlockedDecrement ( & thread_count);

			break;
		case DLL_PROCESS_DETACH:
			DeleteCriticalSection(&gInitCriticalSection);
			break;
	}
	UNREFERENCED_PARAMETER(hinstDll);
	UNREFERENCED_PARAMETER(lpvReserved);
	return(TRUE);
}

