/*
 * Filename:    register.cpp
 *
 * This file handles installation/uninstallation of the UNet spi
 * It's done by DllRegister and UnRegister functions exported by 
 * unetspi.dll, should use regdll32.exe to register it
 *
 * Author: xwh1@cornell.edu
 */
#define UNICODE

#include "unetspi.h"

#define UNET_PROVIDER_NAME L"UNET_PROVIDER"

long                           thread_count ;
HMODULE                        g_hinst ;
CRITICAL_SECTION               gInitCriticalSection;
WSPUPCALLTABLE                 gupCallTable;
INT                            gCatalog;
HANDLE                         hDriver;
UNET_INFO_RESPONSE             hDriverinfo;
INT                            message_size, num_messages;
UNET_CREATE_ENDPOINT_REQUEST   create;
UNET_ACTIVATE_CHANNEL_REQUEST  activate;
UNET_CHANNEL_ID                chan;
UNET_TX_BUFFER_DESCRIPTOR      txbd;
char                           *thebuf;
SocketArray                    socketarray;
INT InstallMyProvider( PDWORD CatalogId)
{
	WSAPROTOCOL_INFOW proto_info;
	int               install_result;
	int               install_error;
	WCHAR             DllPath	[_MAX_PATH];
	char              tempbuf[_MAX_PATH]; tempbuf[0]='\0';
	GUID nullProviderId = {0,0,0,{0,0,0,0,0,0,0,0}};

	// Create a PROTOCOL_INFO to install for our provider DLL.

	// this flag specifies the "service" provided
	// I will set it to be   XP1_MESSAGE_ORIENTED later
	proto_info.dwServiceFlags1 = 0;
	proto_info.dwServiceFlags2 = 0;
	proto_info.dwServiceFlags3 = 0;
	proto_info.dwServiceFlags4 = 0;

	proto_info.dwProviderFlags = PFL_HIDDEN;  // kind of important

	proto_info.ProviderId             = nullProviderId;
	proto_info.dwCatalogEntryId       = 0;   // filled in by system
	proto_info.ProtocolChain.ChainLen = BASE_PROTOCOL; 
	proto_info.iVersion               = 0;
	proto_info.iAddressFamily         = AF_ATM; // important
	proto_info.iMaxSockAddr           = 16;
	proto_info.iMinSockAddr           = 16;
	proto_info.iSocketType            = SOCK_DGRAM;  // important
	proto_info.iProtocol              = IPPROTO_UDP;   // important  
	proto_info.iProtocolMaxOffset     = 0;
	proto_info.iNetworkByteOrder      = BIGENDIAN;
	proto_info.iSecurityScheme        = SECURITY_PROTOCOL_NONE;
	proto_info.dwMessageSize          = 64*1024;  // 64k
	proto_info.dwProviderReserved     = 0xdeadbabe;

	wcscpy( proto_info.szProtocol, UNET_PROVIDER_NAME);

	GetModuleFileName( g_hinst, DllPath, _MAX_PATH);
	WideCharToMultiByte(CP_ACP, 0, DllPath, wcslen ( DllPath) +1 ,
			tempbuf, _MAX_PATH , NULL, NULL);

	install_result = WSCInstallProvider(
			&UNetSPIGuid,
			DllPath,                      // lpszProviderDllPath
			& proto_info,                 // lpProtocolInfoList
			1,                            // dwNumberOfEntries
			& install_error);             // lpErrno

	*CatalogId = proto_info.dwCatalogEntryId;
	return(install_result);
} 

HRESULT DllRegisterServer(void)
{
	// this is the code that is called by regsvr32.exe.
	// so the installation starts here

	WCHAR textBuf[_MAX_PATH];
	DWORD CatalogId;
	INT   RetValue = NOERROR;

	wsprintf(textBuf,
			L"Press OK to contiue installation for UNet Service Provider");
	MessageBox(NULL, textBuf, L"UNet SPI Installation", MB_OK);

	RetValue = InstallMyProvider( & CatalogId);

	return  RetValue;
}

HRESULT DllUnregisterServer(void)
{
	// this is the code that is called by regsvr32.exe for
	// uninstallaion
	// so the uninstallation starts here
	//
	INT RetValue = NOERROR, error;
	RetValue = WSCDeinstallProvider(&UNetSPIGuid, &error) ;
	return RetValue;
}
BOOL WINAPI DllMain(
		IN HINSTANCE hinstDll,
		IN DWORD fdwReason,
		LPVOID lpvReserved
									 )
{

	switch (fdwReason) {

		case DLL_PROCESS_ATTACH:

			// DLL is attaching to the address
			// space of the current process.
			//
			InitializeCriticalSection(&gInitCriticalSection);

			g_hinst      = hinstDll;
			thread_count = 0;
			hDriver      = NULL;

			break;

		case DLL_THREAD_ATTACH:

			// A new thread is being created in the current process.

			InterlockedIncrement ( & thread_count);
			break;

		case DLL_THREAD_DETACH:

			// A thread is exiting cleanly.

			InterlockedDecrement ( & thread_count);
			break;

		case DLL_PROCESS_DETACH:

			// The calling process is detaching
			// the DLL from its address space.
			//
			// Note that lpvReserved will be NULL if the detach is due to
			// a FreeLibrary() call, and non-NULL if the detach is due to
			// process cleanup.
			//

			if( lpvReserved == NULL ) {

			}
			DeleteCriticalSection(&gInitCriticalSection);
			break;
	}
	UNREFERENCED_PARAMETER(hinstDll);
	UNREFERENCED_PARAMETER(lpvReserved);
	return(TRUE);
}

