/*
Module Name:

UNETSPI.CPP : 

Abstract:

	Contains all the entry points for the WS2SPI.  This module implements a
	simple WinSock2 service provider for U-Net. Only UDP is supported

	Sample usage should be refered to ..\test\UDP.

Author:  xwh@cs.cornell.edu

*/
#include "unetspi.h"

#define XWH1_DEBUG
#ifdef XWH1_DEBUG
#define TestPrint printf
#else
#define TestPrint 0 &&
#endif


#define TX_FIFO_SIZE 10

int                      str_to_addr();
HANDLE                   unet_opendev();
struct unet_endpt_user * unet_create_endpt();
UNET_CHANNEL_ID          unet_activate_chan ();
int                      UnetAddr_addr ( UNET_DEV_TYPE devtype,
		sockaddr_unet_fe * feaddr, UNET_CHANNEL_ADDRESS  * addr);



SOCKET
WSPAPI
WSPAccept(
		IN SOCKET s,
		OUT struct sockaddr FAR *addr,
		OUT INT FAR *addrlen,
		IN LPCONDITIONPROC lpfnCondition,
		IN DWORD dwCallbackData,
		OUT INT FAR *lpErrno
				 )
{
	INT          ReturnValue = INVALID_SOCKET;
	TestPrint("Enter WSPAccept()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPAddressToString(
		IN     LPSOCKADDR lpsaAddress,
		IN     DWORD dwAddressLength,
		IN     LPWSAPROTOCOL_INFOW lpProtocolInfo,
		OUT    LPWSTR lpszAddressString,
		IN OUT LPDWORD lpdwAddressStringLength,
		OUT    LPINT lpErrno )
{
	INT                 ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPAddressToString()\n");
	return ReturnValue;
}

INT
WSPAPI
WSPAsyncSelect(
		IN SOCKET s,
		IN HWND hWnd,
		IN unsigned int wMsg,
		IN long lEvent,
		OUT INT FAR *lpErrno
							)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPAsyncSelect()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPBind(
		IN SOCKET s,
		IN const struct sockaddr FAR *name,
		IN INT namelen,
		OUT INT FAR *lpErrno
			 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPBind()\n");
	return(ReturnValue);

}

INT
	WSPAPI
WSPCancelBlockingCall(OUT INT FAR *lpErrno)
{
	return(WSAEINVAL);
}

INT
WSPAPI
WSPCleanup(
		OUT INT FAR *lpErrno
					)
{
	INT          ReturnValue = SOCKET_ERROR;
	TestPrint(" Enter WSPCleanup()\n");

	// close the device
	if ( hDriver ) {
		if ( CloseHandle ( hDriver) == FALSE )
			return SOCKET_ERROR ;
	}
	// perhaps should close the unclosed sockets list as well.  
	// this can be a hell lot faster if maintain a list of 
	// open sockets other than check all of the avail ones //
	for (int i=0;i< 65535;i++)
		if ( socketarray.endpoints[i] !=
				NULL) {
			unet_destroy_endpt ( hDriver,
					socketarray.endpoints[i]->endpointId);
			socketarray.endpoints[i] = NULL;
		}

	TestPrint(" Leave WSPCleanup()");
	return(ReturnValue);
}

INT
WSPAPI
WSPCloseSocket(
		IN SOCKET s,
		OUT INT FAR *lpErrno
							)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPCloseSocket()\n");
	ReturnValue = unet_destroy_endpt ( 
			hDriver, socketarray.endpoints[s]->endpointId);
	socketarray.endpoints[s] = NULL;

	if ( ReturnValue == FALSE)
		return SOCKET_ERROR;
	return(ReturnValue);
}

INT
WSPAPI
WSPConnect(
		IN SOCKET s,
		IN const struct sockaddr FAR *name,
		IN INT namelen,
		IN LPWSABUF lpCallerData,
		IN LPWSABUF lpCalleeData,
		IN LPQOS lpSQOS,
		IN LPQOS lpGQOS,
		OUT INT FAR *lpErrno
					)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter    WSPConnect()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPDuplicateSocket(
		IN SOCKET s,
		IN DWORD dwProcessID,
		IN LPWSAPROTOCOL_INFOW lpProtocolInfo,
		OUT INT FAR *lpErrno
									)
{
	INT ReturnValue=SOCKET_ERROR;
	TestPrint("Enter    WSPDuplicateSocket()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPEnumNetworkEvents(
		IN SOCKET s,
		OUT WSAEVENT hEventObject,
		OUT LPWSANETWORKEVENTS lpNetworkEvents,
		OUT INT FAR *lpErrno
										)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPEnumNetworkEvents()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPEventSelect(
		IN SOCKET s,
		IN OUT WSAEVENT hEventObject,
		IN long lNetworkEvents,
		OUT INT FAR *lpErrno
							)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter  WSPEventSelect()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPGetOverlappedResult(
		IN SOCKET s,
		IN LPWSAOVERLAPPED lpOverlapped,
		IN LPDWORD lpcbTransfer,
		IN BOOL fWait,
		OUT LPDWORD lpdwFlags,
		OUT INT FAR *lpErrno
											)
{
	INT ReturnValue = FALSE;

	TestPrint("Enter WSPGetOverlappedResult()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPGetPeerName(
		IN SOCKET s,
		OUT struct sockaddr FAR *name,
		OUT INT FAR *namelen,
		OUT INT FAR *lpErrno
							)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPGetPeerName()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPGetQOSByName(
		IN SOCKET s,
		IN LPWSABUF lpQOSName,
		IN LPQOS lpQOS,
		OUT INT FAR *lpErrno
							 )
{
	INT ReturnValue = FALSE;
	TestPrint("Enter WSPGetQOSByName()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPGetSockName(
		IN SOCKET s,
		OUT struct sockaddr FAR *name,
		OUT INT FAR *namelen,
		OUT INT FAR *lpErrno
							)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPGetSockName()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPGetSockOpt(
		IN SOCKET s,
		IN INT level,
		IN INT optname,
		OUT char FAR *optval,
		OUT INT FAR *optlen,
		OUT INT FAR *lpErrno
						 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter GetSocketHandle()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPIoctl(
		IN SOCKET s,
		IN DWORD dwIoControlCode,
		IN LPVOID lpvInBuffer,
		IN DWORD cbInBuffer,
		IN LPVOID lpvOutBuffer,
		IN DWORD cbOutBuffer,
		IN LPDWORD lpcbBytesReturned,
		IN LPWSAOVERLAPPED lpOverlapped,
		IN LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine,
		IN LPWSATHREADID lpThreadId,
		OUT INT FAR *lpErrno
				)
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPIoctl()\n");
	return(ReturnValue);
}

SOCKET
WSPAPI
WSPJoinLeaf(
		IN SOCKET s,
		IN const struct sockaddr FAR *name,
		IN INT namelen,
		IN LPWSABUF lpCallerData,
		IN LPWSABUF lpCalleeData,
		IN LPQOS lpSQOS,
		IN LPQOS lpGQOS,
		IN DWORD dwFlags,
		OUT INT FAR *lpErrno
					 )
{
	INT ReturnValue = INVALID_SOCKET;
	TestPrint("Enter WSPJoinLeaf()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPListen(
		IN SOCKET s,
		IN INT backlog,
		OUT INT FAR *lpErrno
				 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPListen()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPRecv(
		IN SOCKET s,
		IN LPWSABUF lpBuffers,
		IN DWORD dwBufferCount,
		IN LPDWORD lpNumberOfBytesRecvd,
		IN OUT LPDWORD lpFlags,
		IN LPWSAOVERLAPPED lpOverlapped,
		IN LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine,
		IN LPWSATHREADID lpThreadId,
		OUT INT FAR *lpErrno
			 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPRecv()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPRecvDisconnect(
		IN SOCKET s,
		IN LPWSABUF lpInboundDisconnectData,
		OUT INT FAR *lpErrno
								 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPRecvDisconnect()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPRecvFrom(
		IN  SOCKET s,
		IN  LPWSABUF lpBuffers,
		IN  DWORD dwBufferCount,
		IN  LPDWORD lpNumberOfBytesRecvd,
		IN  OUT LPDWORD lpFlags,
		OUT struct sockaddr FAR *  lpFrom,
		IN  LPINT lpFromlen,
		IN  LPWSAOVERLAPPED lpOverlapped,
		IN  LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine,
		IN LPWSATHREADID lpThreadId,
		OUT INT FAR *lpErrno
					 )
{
	sockaddr_unet_fe * mysockaddr = ( sockaddr_unet_fe*) lpFrom;
	INT ReturnValue = NO_ERROR;
	TestPrint("Enter WSPRecvFrom()");

	char remoteaddr[]= "00:33:c0:22:39:99.43"; // random
	char localaddr[]= "00:00:c0:43:39:ef.43";

	UNET_ACTIVATE_CHANNEL_REQUEST   activate;
	UNET_DEACTIVATE_CHANNEL_REQUEST deactivate;
	UNET_CHANNEL_ID                 chan;
	struct unet_endpt_user          *endpt;

	endpt = socketarray.endpoints [s];
	activate.endpointId = endpt->endpointId;
	activate.chan       = UNET_CHAN_ANY;

	if ( str_to_addr ( UNET_DEVID_TULIP, remoteaddr, 
			& ( activate. txAddr))){
		TestPrint("Bad address format %s\n", remoteaddr);
		return SOCKET_ERROR;
	}

	if ( UnetAddr_addr(UNET_DEVID_TULIP, mysockaddr, &(activate.rxAddr))){
		TestPrint("Bad address format %s\n", localaddr);
		return SOCKET_ERROR;
	};

	// open channel
	chan = unet_activate_chan ( hDriver, & activate);
	if ( chan  == -1) {
		TestPrint ( "Can't activate channel\n" );
		return SOCKET_ERROR;
	}

	// start to pull data..
	//

	while (endpt->rx_next->chan == UNET_CHANNEL_FREE) ;

	lpBuffers[0].len            = endpt->rx_next->length;
	ReturnValue                 = endpt->rx_next->length;
	*lpNumberOfBytesRecvd       = endpt->rx_next->length;

	if (endpt->rx_next->length > endpt->info.rxd_payload_size) {
		memcpy ( lpBuffers[0].buf, (unsigned char *)
				(endpt->rx_next->pay_or_buf[0] + endpt->buffer_area) ,
				endpt->rx_next->length);

	} else {
		memcpy ( lpBuffers[0].buf , (unsigned char*) 
				endpt->rx_next->pay_or_buf,endpt->rx_next->length );
	}

	if (endpt->rx_next->length > endpt->info.rxd_payload_size) {
		endpt->free_next->buf = (endpt->rx_next->pay_or_buf[0]);
		if (endpt->free_next == endpt->free_last)
			endpt->free_next = endpt->free_first;
		else
			endpt->free_next++;
	}

	endpt->rx_next->chan = UNET_CHANNEL_FREE;
	*(endpt->rx_wall) = ((u_int)endpt->rx_next - (u_int)endpt->rx_first);
	if (endpt->rx_next == endpt->rx_last)
		endpt->rx_next = endpt->rx_first;
	else
		endpt->rx_next++;

	// close channel
	deactivate.endpointId = activate.endpointId;
	deactivate.chan = chan;
	if ( unet_deactivate_chan ( hDriver, & deactivate) == FALSE )
	{
		TestPrint ( " Can't deactivate channel \n");
		return SOCKET_ERROR;
	}

	return(ReturnValue);
}

typedef struct association
{
	SOCKET  ProviderSocket;
	SOCKET  UserSocket;
} SOCKETASSOCIATION, *PSOCKETASSOCIATION;

typedef struct
{
	UINT                AssociationCount;
	PSOCKETASSOCIATION  Associations;
} SOCKETMAP, *PSOCKETMAP;

INT
WSPAPI
WSPSelect(
		IN INT nfds,
		IN OUT fd_set FAR *readfds,
		IN OUT fd_set FAR *writefds,
		IN OUT fd_set FAR *exceptfds,
		IN const struct timeval FAR *timeout,
		OUT INT FAR *lpErrno
				 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPSelect()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPSend(
		IN SOCKET s,
		IN LPWSABUF lpBuffers,
		IN DWORD dwBufferCount,
		IN LPDWORD lpNumberOfBytesSent,
		IN DWORD dwFlags,
		IN LPWSAOVERLAPPED lpOverlapped,
		IN LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine,
		IN LPWSATHREADID lpThreadId,
		OUT INT FAR *lpErrno
			 )
{
	INT              ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPSend()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPSendDisconnect(
		IN SOCKET s,
		IN LPWSABUF lpOutboundDisconnectData,
		OUT INT FAR *lpErrno
								 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPSendDisconnect()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPSendTo(
		IN SOCKET s,
		IN LPWSABUF lpBuffers,
		IN DWORD dwBufferCount,
		IN LPDWORD lpNumberOfBytesSent,
		IN DWORD dwFlags,
		IN const struct sockaddr FAR *  lpTo,
		IN INT iTolen,
		IN LPWSAOVERLAPPED lpOverlapped,
		IN LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine,
		IN LPWSATHREADID lpThreadId,
		OUT INT FAR *lpErrno
				 )
{
	sockaddr_unet_fe * mysockaddr = ( sockaddr_unet_fe*) lpTo;
	INT ReturnValue = NO_ERROR;
	TestPrint("Enter WSPSendTo()\n");

	//
	// should let user pass in adddresss..
	//

	char localaddr[]= "01:00:c0:43:39:ef.00";     //   random.... 
	UNET_TX_BUFFER_DESCRIPTOR       txbd;
	UNET_ACTIVATE_CHANNEL_REQUEST   activate;
	UNET_DEACTIVATE_CHANNEL_REQUEST deactivate;
	UNET_CHANNEL_ID                 chan;
	struct unet_endpt_user          *endpt;

	endpt = socketarray.endpoints [s];

	activate.endpointId = endpt->endpointId;
	activate.chan = UNET_CHAN_ANY;

	if ( UnetAddr_addr ( UNET_DEVID_TULIP, mysockaddr, 
			& ( activate.txAddr))){
		TestPrint("Bad address format \n" );
		return SOCKET_ERROR;
	}

	if ( str_to_addr(UNET_DEVID_TULIP, localaddr, &(activate.rxAddr))){
		TestPrint("Bad address format \n", localaddr);
		return SOCKET_ERROR;
	};

	// open channel
	chan = unet_activate_chan ( hDriver, & activate);
	if ( chan  == -1) {
		TestPrint ( "Can't activate channel\n" );
		return SOCKET_ERROR;
	}

	// start to send
	//

	txbd.buf    = (unsigned long)(lpBuffers[0].buf);
	txbd.length = (unsigned short) lpBuffers[0].len;
	txbd.chan   = chan;

	while ((u_int)endpt->tx_next == 
			((u_int)endpt->tx_first + *(endpt->tx_wall))) {
		UNET_TRAP(hDriver, endpt->endpointId);
	}

	*((UNET_TX_BUFFER_DESCRIPTOR *)(endpt->tx_next)) = txbd;

	UNET_TRAP(hDriver, endpt->endpointId);

	if (endpt->tx_next == endpt->tx_last)
		endpt->tx_next = endpt->tx_first;
	else
		endpt->tx_next++;

	// close channel
	deactivate.endpointId = activate.endpointId;
	deactivate.chan = chan;
	if ( unet_deactivate_chan ( hDriver, & deactivate) == FALSE ) {
		TestPrint ( " Can't deactivate channel \n");
		return SOCKET_ERROR;
	}

	TestPrint("Leave WSPSendTo()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPSetSockOpt(
		IN SOCKET s,
		IN INT level,
		IN INT optname,
		IN const char FAR *optval,
		IN INT optlen,
		OUT INT FAR *lpErrno
						 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPSetSockOpt()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPShutdown(
		IN SOCKET s,
		IN INT how,
		OUT INT FAR *lpErrno
					 )
{
	INT ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPShutdown()\n");
	return(ReturnValue);
}

SOCKET
WSPAPI
WSPSocket(
		IN int af,
		IN int type,
		IN int protocol,
		IN LPWSAPROTOCOL_INFOW lpProtocolInfo,
		IN GROUP g,
		IN DWORD dwFlags,
		OUT INT FAR *lpErrno
				 )
{
	INT    ReturnValue = INVALID_SOCKET;
	int    error;
	// char   addr[]= "00:00:c0:43:39:ef.42";
	UNET_CREATE_ENDPOINT_REQUEST create;
	struct unet_endpt_user * endpt;

	TestPrint("Enter WSPSocket()\n");
	ReturnValue = gupCallTable.lpWPUCreateSocketHandle( gCatalog, 0, & error);

	if ( ReturnValue == INVALID_SOCKET ) {
		TestPrint (" WSPSocket() returning INVALID_SOCKET \n");
		return    ReturnValue;
	} else {
		if ( ReturnValue > 65535) {
			TestPrint (" WSPSocket() returning an invalid Socket handle\n");
			return INVALID_SOCKET;
		}
	}

	// Create an endpoint
	/* XXX The below two should be command-line selectable */
	create.devType = UNET_DEVID_TULIP;
	create.devIndex = 0;
	create.numTxFifoEntries   = TX_FIFO_SIZE;
	create.numFreeFifoEntries = TX_FIFO_SIZE;
	create.numRxFifoEntries   = TX_FIFO_SIZE;
	create.bufferAreaSize     = (TX_FIFO_SIZE+1)*hDriverinfo.max_pdu;
	create.rxBufferSize       = hDriverinfo.max_pdu;

	// create endpoint here
	// socketarray[ReturnValue] 
	endpt = unet_create_endpt (hDriver, &hDriverinfo, &create, TX_FIFO_SIZE);

	if (!endpt) {
		TestPrint("Can't create endpoint!\n");
		return INVALID_SOCKET;
	}

	EnterCriticalSection (& gInitCriticalSection);

	socketarray.endpoints[ReturnValue]  = endpt;

	LeaveCriticalSection ( &gInitCriticalSection);


	TestPrint ( "Done in WSPSocket\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPStringToAddress(
		IN     LPWSTR AddressString,
		IN     INT AddressFamily,
		IN     LPWSAPROTOCOL_INFOW lpProtocolInfo,
		OUT    LPSOCKADDR lpAddress,
		IN OUT LPINT lpAddressLength,
		IN OUT LPINT lpErrno )
{
	INT                 ReturnValue = SOCKET_ERROR;
	TestPrint("Enter WSPStringToAddress()\n");
	return(ReturnValue);
}

INT
WSPAPI
WSPStartup(
		WORD wVersion,
		LPWSPDATA lpWSPData,
		LPWSAPROTOCOL_INFOW lpProtocolInfo,
		WSPUPCALLTABLE UpcallTable,
		LPWSPPROC_TABLE lpProcTable )
{
	TestPrint("Enter WSPStartup()\n" ) ;
	INT ReturnCode;

	EnterCriticalSection(&gInitCriticalSection);

	ReturnCode   = NO_ERROR;
	gupCallTable = UpcallTable ;
	gCatalog     = lpProtocolInfo->dwCatalogEntryId;

	LeaveCriticalSection(&gInitCriticalSection);

	TestPrint ( " MY Catalog = %d\n", gCatalog );


	lpProcTable->lpWSPAccept              = WSPAccept;
	lpProcTable->lpWSPAddressToString     = WSPAddressToString;
	lpProcTable->lpWSPAsyncSelect         = WSPAsyncSelect;
	lpProcTable->lpWSPBind                = WSPBind;
	lpProcTable->lpWSPCancelBlockingCall  = WSPCancelBlockingCall;
	lpProcTable->lpWSPCleanup             = WSPCleanup;
	lpProcTable->lpWSPCloseSocket         = WSPCloseSocket;
	lpProcTable->lpWSPConnect             = WSPConnect;
	lpProcTable->lpWSPDuplicateSocket     = WSPDuplicateSocket;
	lpProcTable->lpWSPEnumNetworkEvents   = WSPEnumNetworkEvents;
	lpProcTable->lpWSPEventSelect         = WSPEventSelect;
	lpProcTable->lpWSPGetOverlappedResult = WSPGetOverlappedResult;
	lpProcTable->lpWSPGetPeerName         = WSPGetPeerName;
	lpProcTable->lpWSPGetSockName         = WSPGetSockName;
	lpProcTable->lpWSPGetSockOpt          = WSPGetSockOpt;
	lpProcTable->lpWSPGetQOSByName        = WSPGetQOSByName;
	lpProcTable->lpWSPIoctl               = WSPIoctl;
	lpProcTable->lpWSPJoinLeaf            = WSPJoinLeaf;
	lpProcTable->lpWSPListen              = WSPListen;
	lpProcTable->lpWSPRecv                = WSPRecv;
	lpProcTable->lpWSPRecvDisconnect      = WSPRecvDisconnect;
	lpProcTable->lpWSPRecvFrom            = WSPRecvFrom;
	lpProcTable->lpWSPSelect              = WSPSelect;
	lpProcTable->lpWSPSend                = WSPSend;
	lpProcTable->lpWSPSendDisconnect      = WSPSendDisconnect;
	lpProcTable->lpWSPSendTo              = WSPSendTo;
	lpProcTable->lpWSPSetSockOpt          = WSPSetSockOpt;
	lpProcTable->lpWSPShutdown            = WSPShutdown;
	lpProcTable->lpWSPSocket              = WSPSocket;
	lpProcTable->lpWSPStringToAddress     = WSPStringToAddress;

	// Set the version info
	lpWSPData->wVersion     = MAKEWORD(2,2); 
	lpWSPData->wHighVersion = MAKEWORD(2,2);

	return(ReturnCode);
}
