/*
 * Copyright (c) 1996 by Cornell University.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 */

#ifndef UDI_H
#define UDI_H

// U-Net Adapter interface
//
// The following functions form the interface between a U-Net adapter and
// the UDI. Every adapter provides its set of these funtions and registers
// then with the UDI using the object returned by UnetCreateAdapterObject below.

typedef
NTSTATUS
(*PUNET_ADAPTER_CREATE_ENDPOINT) (
    IN UNET_ENDPOINT_ID UnetEndpointId,
    IN PUNET_CREATE_ENDPOINT_REQUEST UnetCreateEndpointRequest,
    OUT PUNET_CREATE_ENDPOINT_RESPONSE UnetCreateEndpointResponse
    );

typedef
NTSTATUS
(*PUNET_ADAPTER_DESTROY_ENDPOINT) (
    IN UNET_ENDPOINT_ID UnetEndpointId
    );
    
typedef
NTSTATUS
(*PUNET_ADAPTER_ACTIVATE_CHANNEL) (
    IN PUNET_ACTIVATE_CHANNEL_REQUEST UnetActivateChannelRequest,
    IN PUNET_ACTIVATE_CHANNEL_RESPONSE UnetActivateChannelResponse
    );
    
typedef
NTSTATUS
(*PUNET_ADAPTER_DEACTIVATE_CHANNEL) (
    IN PUNET_DEACTIVATE_CHANNEL_REQUEST UnetDeactivateChannelRequest
    );
    
typedef
NTSTATUS
(*PUNET_ADAPTER_TRANSMIT) (
													 IN UNET_ENDPOINT_ID UnetEndpointId
													 );
    
typedef
VOID
(*PUNET_ADAPTER_STOP) (
    );

typedef
NTSTATUS
(*PUNET_ADAPTER_INFO) (
											 OUT PUNET_INFO_RESPONSE UnetInfoRequest
											 );

typedef
NTSTATUS
(*PUNET_ADAPTER_IOCTL) (
											 IN PUNET_ADAPTER_IOCTL_REQUEST UnetAdapterIoctlRequest,
											 OUT PUNET_ADAPTER_IOCTL_RESPONSE UnetAdapterIoctlResponse
											 );

typedef struct _UNET_ADAPTER *PUNET_ADAPTER;
typedef struct _UNET_ADAPTER {
	PUNET_ADAPTER					nextAdapter;
	UNET_DEV_TYPE			devType;
	UNET_DEV_INDEX			devIndex;
	PUNET_ADAPTER_CREATE_ENDPOINT	createEndpoint;
	PUNET_ADAPTER_DESTROY_ENDPOINT	destroyEndpoint;
	PUNET_ADAPTER_ACTIVATE_CHANNEL	activateChannel;
	PUNET_ADAPTER_DEACTIVATE_CHANNEL deactivateChannel;
	PUNET_ADAPTER_INFO				adapterInfo;
	PUNET_ADAPTER_TRANSMIT			transmit;
	PUNET_ADAPTER_STOP				stopAdapter;
	PUNET_ADAPTER_IOCTL       adapterIoctl;
} UNET_ADAPTER;

typedef struct {
	ULONG               state;
	UNET_ENDPOINT_ID    endpointId;
	PUNET_ADAPTER       adapter;
	PFILE_OBJECT        fileObject;
} UNET_ENDPOINT, *PUNET_ENDPOINT;

PUNET_ADAPTER
UnetCreateAdapterObject(
												UNET_DEV_TYPE UnetDevType,
												UNET_DEV_INDEX UnetDevIndex
												);

NTSTATUS
UnetDestroyAdapterObject(
												 UNET_DEV_TYPE UnetDevType,
												 UNET_DEV_INDEX UnetDevIndex
												 );

PUNET_ADAPTER
UnetFindAdapterObject(
											UNET_DEV_TYPE UnetDevType,
											UNET_DEV_INDEX UnetDevIndex
											);

// General UDI parameters

#define UNET_DEFAULT_MAX_ENDPOINTS          10
#define UNET_DEFAULT_MAX_COMSEG_SIZE        1024*1024
#define UNET_DEFAULT_MAX_TOTAL_COMSEG_SIZE  10*1024*1024
#define UNET_DEFAULT_COMSEG_SIZE            512*1024
#define UNET_DEFAULT_TX_FIFO_SIZE           255
#define UNET_DEFAULT_RX_FIFO_SIZE           255
#define UNET_DEFAULT_FREE_FIFO_SIZE         512
#define UNET_DEFAULT_RX_BUFFER_SIZE         1024

#define UNET_ENDPOINT_IDLE  0
#define UNET_ENDPOINT_INUSE 1

#define UNET_DOSDEVICE_NAME L"\\DosDevices\\UNET"
#define UNET_DEVICE_NAME L"\\Device\\Unet"
    
typedef struct {
	ULONG   maxEndpoints;
	ULONG   maxComSegSz;
	ULONG   maxTotalComSegSz;
	ULONG   defaultComSegSz;
	ULONG   defaultTxFifoSz;
	ULONG   defaultRxFifoSz;
	ULONG   defaultFreeFifoSz;
	ULONG   defaultRxBufferSz;
	UNET_ENDPOINT *endpoints;
} UNET_OBJECT, *PUNET_OBJECT;

typedef NTSTATUS
(*PUNET_TRAP_FUNCTION) (
												IN ULONG    TrapArgument
												);

NTSTATUS
NtUnetTrapRegister(
									 PUNET_TRAP_FUNCTION TrapFunction
									 );

NTSTATUS
NtUnetTrapDeRegister();

#endif UDI_H

