
/***********************************************/
/* Server Test for U-Net NT over Fast Ethernet */
/* by So, Hoi Sheung <so@cs.cornell.edu>       */
/* U-Net Group, Computer Science,              */
/* Cornell University                          */
/* All rights reserved                         */
/* Last Modified: 11 Feb, 97.                  */
/***********************************************/
/********************************************************************************************/
/* servertest simulates a Client/Server system.  The Client sends a request to the Server   */
/* The Server returns the specified number of packets.  The Client times the roundtrip time */
/* and throughput;  the Server times the sending speed                                      */
/********************************************************************************************/

/********************************************************************************************/
/* Note: In order to time the operations accurately on a Pentium/Pentium Pro Machine,       */
/*       please set the MHZ definition in libunet.h to the clock rate of your Pentium/PPro  */
/********************************************************************************************/

#define TX_FIFO_SIZE 100
#define EMPTY_TRAP_LIMIT 10000000 /* no. of empty traps before we realize a pkt has been lost */
//#define HOI_DEBUG
//#define VERBOSE_DEBUG
//#define SLEEP  /* calls system command sleep() and runs process scheduler */
//#define NAP {long int k; for(k=0; k<200; k++);}   /* empty loop */
//#define TIME_TEST

#define MUTE
#define SLEEP_TIME 1  /* millisecond */
#define PORT1 "11"    
#define PORT2 "22"     

#include <windows.h>
#include <winioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "unetuser.h"
#include "libunet.h"


unsigned char sendbuf[4096];

int main(int argc, char* argv[]) {
	
	char input[80];
	char localaddr[25];  /* Ethernet address w/ portnumber */
	char testtype;

	HANDLE hDriver; /* handle of the device driver */
	UNET_INFO_RESPONSE info; /* response of unet device driver */
	int message_size, num_messages; /* size & num of msgs to be sent or received*/
	struct unet_endpt_user *endpt; /* ptr to endpoint */
	UNET_CREATE_ENDPOINT_REQUEST create; /* info to pass to unet to create endpoint */
	UNET_ACTIVATE_CHANNEL_REQUEST activate; /* info to activate channel */
	UNET_DEACTIVATE_CHANNEL_REQUEST deactivate;
	UNET_CHANNEL_ID chan;
	UNET_TX_BUFFER_DESCRIPTOR txbd;
	double t1, t2, t3; /* time */
	int i, j;

	printf("\n\nU-Net Sample Client/Server Program V1.0 Feb 1997.\n");
#ifdef TIME_TEST
	t1 = get_seconds();
	for(i=0; i<1\98; i++) 
		t2 = get_seconds();
	t2 = get_seconds();
	printf("Time for 102 get_sec() = %20.15f us\n", (t2-t1)/100*(1e6));
#endif
	/***************************************************************/
	/* open U-Net device *******************************************/
	/***************************************************************/
	hDriver = unet_opendev(UNET_DEVID_TULIP, 0, &info);
	if (!hDriver) {
		printf("Error: Can't open U-Net device\n");
		exit(1);
	}
	/***************************************************************/
	/* find out the Ethernet address of this device                */
	/***************************************************************/
	for (i=0;i<6;i++)
		sprintf((localaddr+i*3), "%02x%c", info.local_addr.fe.macAddr[i], (i==5)?'.':':');



	/* Ask whether user want to send or receive */
	do {
		printf("Is this a (s)erver or a (c)lient?\n");
		gets(input);
		testtype = input[0];
		if (testtype == 's') testtype = 'S';
		if (testtype == 'c') testtype = 'C';
	} while (testtype!='S' && testtype!='C');

#ifdef HOI_DEBUG
    /* if sending between SYSLAB05 & SYSLAB06 */
	if (testtype=='S') {
		str_to_addr(UNET_DEVID_TULIP, "00:00:c0:39:71:9c.10", &(activate.rxAddr));
		str_to_addr(UNET_DEVID_TULIP, "00:00:c0:17:bc:c7.20", &(activate.txAddr));
	} else { /* this is a recv station */
		str_to_addr(UNET_DEVID_TULIP, "00:00:c0:17:bc:c7.20", &(activate.rxAddr));
		str_to_addr(UNET_DEVID_TULIP, "00:00:c0:39:71:9c.10", &(activate.txAddr));
	} 
#endif 

	/* Ask for addresses */
	if (testtype=='S') {
#ifndef HOI_DEBUG
		/* local address & listening port */
		do {
			printf("Enter Ethernet Addr & Port Num of SERVER(this computer)?\n");
			printf("Default = %s%s\n", localaddr, PORT1);
			gets(input);
			if (strlen(input)==0) {
				strcpy(input, localaddr);
				strcat(input, PORT1);
			}
		} while (str_to_addr(UNET_DEVID_TULIP, input, &(activate.rxAddr)));

		/* remote address & port */
		do {
			printf("Enter Ethernet Addr & Port Num of CLIENT?\n");
			printf("Default = %s%s\n", localaddr, PORT2);
			gets(input);
			if (strlen(input)==0) {
				strcpy(input, localaddr);
				strcat(input, PORT2);
			}
		} while (str_to_addr(UNET_DEVID_TULIP, input, &(activate.txAddr)));
#endif
	} else { /* this is a recv station */
#ifndef HOI_DEBUG
		/* local address & listening port */
		do {
			printf("Enter Ethernet Addr & Port Num of CLIENT(this computer)?\n");
			printf("Default = %s%s\n", localaddr, PORT2);
			gets(input);
			if (strlen(input)==0) {
				strcpy(input, localaddr);
				strcat(input, PORT2);
			}
		} while (str_to_addr(UNET_DEVID_TULIP, input, &(activate.rxAddr)));

		/* remote address & port */
		do {
			printf("Enter Ethernet Addr & Port Num of SERVER?\n");
			printf("Default = %s%s\n", localaddr, PORT1);
			gets(input);
			if (strlen(input)==0) {
				strcpy(input, localaddr);
				strcat(input, PORT1);
			}
		} while (str_to_addr(UNET_DEVID_TULIP, input, &(activate.txAddr)));
#endif

		/* ask for num and size of messages ****************************/
		do {
			printf("Enter number of messages to receive?\n");
			gets(input);
			num_messages = atoi(input);
		} while (num_messages<=0);
		do {
			printf("Enter size of each message?\n");
			gets(input);
			message_size = atoi(input);
		} while (message_size<=0 || message_size>info.max_pdu);
	} 

    /***************************************************************/
	/* create an end point *****************************************/
	/***************************************************************/
	create.devType = UNET_DEVID_TULIP;
	create.devIndex = 0;  /* the first tulip board */
	create.numTxFifoEntries = TX_FIFO_SIZE;
	create.numFreeFifoEntries = TX_FIFO_SIZE;
	create.numRxFifoEntries = TX_FIFO_SIZE;
	// create.bufferAreaSize = (TX_FIFO_SIZE+1)*info.max_pdu;
	create.bufferAreaSize = 0;
	create.rxBufferSize = info.max_pdu;

	endpt = unet_create_endpt(hDriver, &info, &create, TX_FIFO_SIZE);
	/* memory has been allocated by unet device driver */

	if (!endpt) {
		printf("Error: Can't create endpoint\n");
		exit(1);
	}
    /***************************************************************/
	/* open a channel **********************************************/
    /***************************************************************/
	activate.endpointId = endpt->endpointId;
	i = 1;
	/* choose a channel if a channel is alread occupied */
	do {
		activate.chan = i;
		chan = unet_activate_chan(hDriver, &activate);
		i++;
	} while (i<256 && chan == -1);

	if (chan == -1) {
		printf("Error: Can't open channel.\n");
		exit(1);
	} else {
		printf("Channel #%d openned\n",chan);
	}


    /***************************************************************/
	/* Server Side *************************************************/
    /***************************************************************/

	if (testtype == 'S') {
		int i;
		UNET_TX_BUFFER_DESCRIPTOR* final;
		unsigned char *incoming;
		int len;  /* length of received message */	
		int num_bufs; /* number of buffers used to told */

		/************************************************************/
		/* waiting for a request ************************************/
		/************************************************************/
	    while (endpt->rx_next->chan == UNET_CHANNEL_FREE) ;
		len = endpt->rx_next->length; /* length of received message */
		num_bufs = (len / endpt->rxb_buffsize_used) + 1;
		incoming = (unsigned char *) endpt->rx_next->pay_or_buf;
		num_messages = (*(int *)incoming);
		message_size = (* ((int *)incoming + 1));
		printf("Received a request for %d %d-Byte messages\n", num_messages, message_size);

		/* generate the message to be sent ****************************/
		for ( i = 0; i < message_size; i++) {
			sendbuf[i] = (unsigned char) i&0xFF;
		}

		/************************************************************/
		/* setup the transmission buf descriptor ********************/
		/************************************************************/
		txbd.buf = (unsigned long) sendbuf;
		txbd.length = message_size;
		txbd.chan = chan;

		t1 = get_seconds();

		/* now sending messages */
		for (i=0; i < num_messages; i++) {
			/* wait until there is a space in the tx queue */
			while ((u_int)endpt->tx_next == ((u_int)endpt->tx_first + *(endpt->tx_wall))) {
				UNET_TRAP(hDriver, endpt->endpointId);
			}

			/* push the message into  the tx queue */
			*(endpt->tx_next) = txbd;

			/* trap into kernel to send */
			//UNET_TRAP(hDriver, endpt->endpointId);
#ifdef SLEEP
			Sleep(SLEEP_TIME);
#endif

			/* store the pointer to the last descriptor */
			/* it will be used later to make sure all packets */
			/* have been serviced */
			final = endpt->tx_next;
		

			/* increment tx next */
			if (endpt->tx_next == endpt->tx_last)
				endpt->tx_next = endpt->tx_first;
			else
				endpt->tx_next++;
		}	

		/************************************************************/
		/* Keep calling UNet to send until last pkt has been sent ***/
		/************************************************************/
		while (final->chan==chan) {
			UNET_TRAP(hDriver, endpt->endpointId);
		}
				
		t2 = get_seconds();
		printf("\nSent %d %d-byte message in %f ms \n\n",
			num_messages, message_size, (t2-t1)*1000);
		printf("Throughput = %f us/msg or %f Mbps\n\n",
			((t2-t1)*1e6)/num_messages,
			(num_messages*message_size*8)/(1e6*(t2-t1)));


	} else {  
    /***************************************************************/
	/* Client Side *************************************************/
    /***************************************************************/

		int len;  /* length of received message */	
		int num_bufs; /* number of buffers used to told */
		int i; 
		long int empty_trap;
		/**************************************************************/
		/* send a message telling the dest the num of messages coming */
		/**************************************************************/
		*(int *)sendbuf = num_messages;
		*((int *)sendbuf + 1) = message_size;
		txbd.buf = (unsigned long) sendbuf;
		txbd.length = sizeof(int) * 2;
		txbd.chan = chan;
		while ((u_int)endpt->tx_next == ((u_int)endpt->tx_first + *(endpt->tx_wall)))
			UNET_TRAP(hDriver, endpt->endpointId);
		*(endpt->tx_next) = txbd;
		UNET_TRAP(hDriver, endpt->endpointId);
		t1 = get_seconds();

		if (endpt->tx_next == endpt->tx_last)
			endpt->tx_next = endpt->tx_first;
		else
			endpt->tx_next++;

		/************************************/
		/* receiving packets from the server*/
		/************************************/
		for (i=0; i<num_messages; i++) {
            empty_trap = 0;
		    while (endpt->rx_next->chan == UNET_CHANNEL_FREE) {
				if (++empty_trap > EMPTY_TRAP_LIMIT) {
					printf("%d packets received, %d packets lost\n", i, num_messages-i);
				    exit(1);
				}
			}


			/* time when first message was received */
			if (i==0) t2 = get_seconds();

			len = endpt->rx_next->length; /* length of received message */
			num_bufs = (len / endpt->rxb_buffsize_used) + 1;


			if (endpt->rx_next->length > endpt->info.rxd_payload_size) {
				/*********************************************/
				/* packet data do not fit in 1 descriptor    */
				/*********************************************/
#ifndef MUTE
				printf("Received buffered message #%d of length %d\n", i, len);		
#endif
				for (j = 0; j < num_bufs; j++) {
#ifdef VERBOSE_DEBUG
					/* print out the content of imcoming message */
					unsigned char *incoming;
					unsigned int k;
					incoming = (unsigned char *) endpt->rx_next->pay_or_buf[j];
					for (k=0; k<len; k++) {  
						printf("%2X",(unsigned char) (*(incoming+k)));
					}
					printf("\n");
#endif				
					/***************************************************/
				    /*  push the buffer space back to the free queue   */
					/***************************************************/
					endpt->free_next->buf = (endpt->rx_next->pay_or_buf[j]);
					/* increment free_next */
					if (endpt->free_next == endpt->free_last)
						endpt->free_next = endpt->free_first;
					else
						endpt->free_next++;
				}
			} else {
				/**************************************/
				/* packet data do fit in 1 descriptor */
				/**************************************/
#ifndef MUTE
				printf("Received inline message #%d of length %d\n", i, len);		
#endif
#ifdef VERBOSE_DEBUG
				unsigned char *incoming;
				unsigned int k;
				incoming = (unsigned char *) endpt->rx_next->pay_or_buf;
				for (k=0; k<len; k++) {  
					printf("%2X",(unsigned char) (*(incoming+k)));
				}
				printf("\n");
#endif
			}
			/******************/
			/* update rx fifo */
			/******************/
			endpt->rx_next->chan = UNET_CHANNEL_FREE;
			*(endpt->rx_wall) = ((u_int)endpt->rx_next - (u_int)endpt->rx_first);			
			if (endpt->rx_next == endpt->rx_last)
				endpt->rx_next = endpt->rx_first;
			else
				endpt->rx_next++;
		}
		t3 = get_seconds();
		printf("\nR/T time from request to receiving fisrt pkt = %f ms\n\n", (t2-t1)*1000);

		printf("Total time taken = %f ms\n\n", (t3-t1)*1000);

		printf("Throughput = %f usec/msg or %f Mbps\n\n", 
			((t3-t1)*1e6)/num_messages,
			(num_messages*message_size*8)/(1e6*(t3-t1)));
	}

    /***************************************************************/
	/* deactivate channel, destroy endpoint & close device driver  */
    /***************************************************************/
	deactivate.endpointId = endpt->endpointId;
	deactivate.chan = chan;
	unet_deactivate_chan(hDriver, &deactivate);
	unet_destroy_endpt(hDriver, endpt->endpointId);
	CloseHandle(hDriver);		

	return 0;
}



	